/*
 *	HT Editor
 *	htvfs.cc
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include "htctrl.h"
#include "htiobox.h"
#include "htkeyb.h"
#include "htvfs.h"
#include "httag.h"
#include "htstring.h"

/*
 *	CLASS ht_vfs
 */

int ht_vfs::canonicalize(char *in_name, char *out_name)
{
	return ENOSYS;
}

int ht_vfs::delete_file(char *filename)
{
	return ENOSYS;
}

int ht_vfs::filename_compare(char *a, char *b)
{
	return 0;
}

bool ht_vfs::findclose(pfind_t *f)
{
	return false;
}

bool ht_vfs::findfirst(char *dirname, pfind_t *f)
{
	return false;
}

bool ht_vfs::findnext(pfind_t *f)
{
	return false;
}

int ht_vfs::get_caps()
{
	return 0;
}

int ht_vfs::mkdir(char *dirname)
{
	return ENOSYS;
}

int ht_vfs::open(char *filename, bool edit)
{
	return ENOSYS;
}

int ht_vfs::pstat(pstat_t *s, char *filename)
{
	return ENOSYS;
}

int ht_vfs::rename(char *filename, char *newname)
{
	return ENOSYS;
}

bool ht_vfs::set_protocol(char *protocol)
{
	return 0;
}
	
/*
 *	CLASS ht_file_vfs
 */

void ht_file_vfs::init()
{
	ht_vfs::init();
}

void ht_file_vfs::done()
{
	ht_vfs::done();
}

int ht_file_vfs::canonicalize(char *in_name, char *out_name)
{
	return sys_canonicalize(in_name, out_name);
}

int ht_file_vfs::delete_file(char *filename)
{
	char out_name[260];
	if (canonicalize(filename, out_name)!=0) return ENOENT;
	return remove(out_name);
}

int ht_file_vfs::filename_compare(char *a, char *b)
{
/* FIXME: backslash & slash */
	if (strcmp(a, "..")==0) return -1;
	if (strcmp(b, "..")==0) return 1;
	return ht_stricmp(a, b);
}

bool ht_file_vfs::findfirst(char *dirname, pfind_t *f)
{
	return (sys_findfirst(dirname, f)==0);
}

bool ht_file_vfs::findnext(pfind_t *f)
{
	return (sys_findnext(f)==0);
}

bool ht_file_vfs::findclose(pfind_t *f)
{
	return (sys_findclose(f)==0);
}

int ht_file_vfs::get_caps()
{
	return VFSCAP_WRITABLE;
}

int ht_file_vfs::open(char *filename, bool edit)
{
	return ENOSYS;
}

int ht_file_vfs::pstat(pstat_t *s, char *filename)
{
	return sys_pstat(s, filename);
}

bool ht_file_vfs::set_protocol(char *protocol)
{
	return ((!*protocol) || (!strcmp(protocol, "file")));
}
	
/*
 *	CLASS ht_reg_vfs
 */

void ht_reg_vfs::init(ht_registry *_registry)
{
	ht_vfs::init();
	registry=_registry;
	enum_second=0;
}

void ht_reg_vfs::done()
{
	ht_vfs::done();
}

int ht_reg_vfs::canonicalize(char *in_name, char *out_name)
{
	ht_registry_data *data;
	ht_registry_node_type type;
	strcpy(out_name, in_name);
	int ln=strlen(out_name);
	if ((ln>=3) && (strcmp(out_name+ln-3, "/..")==0)) {
		out_name[ln-3]=0;
		char *last_slash = strrchr(out_name, '/');
		if (!last_slash) return ENOENT;
		if (last_slash==out_name) last_slash++;
		*last_slash=0;
	}
	return registry->find_data_entry(out_name, &data, &type, 0);
}

void ht_reg_vfs::create_pfind_t(pfind_t *f, char *key, ht_registry_data *data, ht_registry_node_type type)
{
	f->name=key;
	f->stat.caps=pstat_mode_type | pstat_desc;
	f->stat.mode=0;
	switch (type) {
		case RNT_SUBDIR:
			f->stat.mode|=HT_S_IFDIR;
			break;
		case RNT_SYMLINK:
			f->stat.mode|=HT_S_IFLNK;
			break;
		default:
			f->stat.mode|=HT_S_IFREG;
	}
	data->strvalue(f->stat.desc);		/* FIXME: possible buffer overflow !!! only 32 bytes... */
}

int ht_reg_vfs::delete_file(char *filename)
{
	return registry->delete_node(filename) ? 0 : ENOSYS;
}

int ht_reg_vfs::filename_compare(char *a, char *b)
{
	return strcmp(a, b);
}

bool ht_reg_vfs::findfirst(char *dirname, pfind_t *f)
{
	char *key;
	ht_registry_data *data;
	ht_registry_node_type type;
	
	if (strcmp(dirname, "/")==0) {
		if (registry->enum_first(dirname, &key, &data, &type)) {
			if (enum_second) free(enum_second);
			enum_second=0;
			create_pfind_t(f, key, data, type);
			return 1;
		}
	} else {
		enum_second=strdup(dirname);
		f->name="..";
		f->stat.caps=pstat_mode_type;
		f->stat.mode=HT_S_IFDIR;
		return 1;
	}
	return 0;
}

bool ht_reg_vfs::findnext(pfind_t *f)
{
	char *key;
	ht_registry_data *data;
	ht_registry_node_type type;
	
	if (enum_second) {
		if (registry->enum_first(enum_second, &key, &data, &type)) {
			if (enum_second) free(enum_second);
			enum_second=0;
			create_pfind_t(f, key, data, type);
			return 1;
		}
		if (enum_second) free(enum_second);
		enum_second=0;
	} else {
		if (registry->enum_next(&key, &data, &type)) {
			create_pfind_t(f, key, data, type);
			return 1;
		}
	}
	return 0;
}

bool ht_reg_vfs::findclose(pfind_t *f)
{
	return 1;
}

int ht_reg_vfs::get_caps()
{
	return VFSCAP_WRITABLE;
//	return 0;
}

int ht_reg_vfs::mkdir(char *dirname)
{
	return !(registry->create_subdir(dirname));
}

int ht_reg_vfs::open(char *filename, bool edit)
{
	ht_registry_data *data;
	ht_registry_node_type type;
	
	if (registry->find_data_entry(filename, &data, &type, 0)) {
		if (data->editdialog(filename)) {
			htmsg m;
			m.msg=msg_config_changed;
			m.type=mt_broadcast;
			app->sendmsg(&m);
		}
		return 0;
	}
	return ENOSYS;
}

int ht_reg_vfs::pstat(pstat_t *s, char *filename)
{
	return ENOSYS;
}

bool ht_reg_vfs::set_protocol(char *protocol)
{
	return (!strcmp(protocol, "reg"));
}

/*
 *	CLASS ht_vfs_viewer
 */
 
void ht_vfs_viewer::init(bounds *b, char *desc, int caps, ht_streamfile *file, ht_format_group *format_group, ht_vfs_viewer_status *s)
{
	ht_uformat_viewer::init(b, desc, caps, file, format_group);
	status=s;
}

void ht_vfs_viewer::config_changed()
{
	ht_uformat_viewer::config_changed();
/* FIXME: refreshdir() shouldn't be public, we should use messaging when available for ht_sub */
	if (cursor_sub) ((ht_vfs_sub*)cursor_sub)->refreshdir();
}

char *ht_vfs_viewer::func(UINT i, bool execute)
{
	if (cursor_sub) {
		ht_vfs_sub *s=(ht_vfs_sub*)cursor_sub;
/* FIXME: cvfs shouldn't be public, we should use messaging when available for ht_sub */
		int caps=s->cvfs->get_caps();
		switch (i) {
			case 7:
				if (caps & VFSCAP_WRITABLE) {
					if (execute) {
						char dir[261] = "";
						if (inputbox("make directory", "new directory: ", dir, 260) == button_ok) {
							s->mkdir(dir);
							dirtyview();
						}
					}
					return "mkdir";
				} else {
					return "~mkdir";
				}
			case 8:
				if (caps & VFSCAP_WRITABLE) {
					if (execute) {
						s->unlink(cursor_id1, cursor_id2);
						dirtyview();
					}
					return "delete";
				} else {
					return "~delete";
				}
				break;
		}
	}
	return ht_uformat_viewer::func(i, execute);
}

void ht_vfs_viewer::handlemsg(htmsg *msg)
{
	switch (msg->msg) {
		case msg_complete_init:
			ht_uformat_viewer::handlemsg(msg);
			update_status();
			return;
		case msg_keypressed:
			switch (msg->data1.integer) {
				case K_Return:
					ref();
					clearmsg(msg);
					return;
			}
			break;
	}
	return ht_uformat_viewer::handlemsg(msg);
}

int ht_vfs_viewer::ref()
{
	int r=ht_uformat_viewer::ref();
	update_status();
	return r;
}

void ht_vfs_viewer::update_status()
{
/* FIXME: cwd and cproto shouldn't be public, we should use messaging when available for ht_sub */
	if (cursor_sub) {
		ht_vfs_viewer_status_data d;
		d.cproto=((ht_vfs_sub*)cursor_sub)->cproto;
		d.cwd=((ht_vfs_sub*)cursor_sub)->cwd;
		if (status) status->setdata(&d);
	}
}

/*
 *	CLASS ht_vfs_viewer_status
 */
 
void ht_vfs_viewer_status::init(bounds *b)
{
	ht_group::init(b, 0, 0);

	b->y=0;
	b->h=1;

	b->x=10;
	if (b->w<20) b->w=0; else b->w-=20;
	t=new ht_statictext();
	t->init(b, "?", align_left);
	   
	b->x=0;
	b->w=10;
	l=new ht_label;
	l->init(b, "~jo", t);

//	insert(l);
	insert(t);
}

void ht_vfs_viewer_status::setdata(void *buf)
{
	ht_vfs_viewer_status_data *d=(ht_vfs_viewer_status_data*)buf;
	t->settext(d->cwd);
	t->dirtyview();
}
			 
/*
 *	CLASS ht_vfs_sub
 */

#define VFSV_FORMAT_NAME		0
#define VFSV_FORMAT_SIZE		1
#define VFSV_FORMAT_BSIZE	2
#define VFSV_FORMAT_TYPE		3
#define VFSV_FORMAT_MTIME	4
#define VFSV_FORMAT_ATIME	5
#define VFSV_FORMAT_CTIME	6
#define VFSV_FORMAT_PERM		7
#define VFSV_FORMAT_MODE		8
#define VFSV_FORMAT_NLINK	9
#define VFSV_FORMAT_NGID		10
#define VFSV_FORMAT_NUID		11
#define VFSV_FORMAT_OWNER	12
#define VFSV_FORMAT_GROUP	13
#define VFSV_FORMAT_INODE	14
#define VFSV_FORMAT_SPACE	15
#define VFSV_FORMAT_MARK		16
#define VFSV_FORMAT_SEPARATOR	17
#define VFSV_FORMAT_DESC		18
#define VFSV_FORMAT_RMTIME	19
#define VFSV_FORMAT_RATIME	20
#define VFSV_FORMAT_RCTIME	21

char *format_property[VFSV_FORMAT_PROPERTIES]={
	"name",	"size",
	"bsize",	"type",
	"mtime",	"atime",
	"ctime",	"perm",
	"mode",	"nlink",
	"ngid",	"nuid",
	"owner",	"group",
	"inode",	"space",
	"mark",	"|",
	"desc",	"rmtime",
	"ratime",	"rctime"
};

#define MAKE_DISPLAY_FORMAT(type) MAKE_DISPLAY_FORMAT_MIN_WIDTH(type, 0)
#define MAKE_DISPLAY_FORMAT_MIN_WIDTH(type, width) (((width) << 16) | (type))
#define MAKE_DISPLAY_FORMAT_FIXED_WIDTH(type, width) ((((width) | 0x8000) << 16) | (type))

#define GET_DISPLAY_FORMAT_TYPE(dfmt) ((dfmt) & 0xffff)
#define GET_DISPLAY_FORMAT_SIZE(dfmt) (((dfmt)>>16) & 0x7fff)
#define GET_DISPLAY_FORMAT_IS_FIXED_SIZE(dfmt) ((dfmt) & 0x80000000)
#define GET_DISPLAY_FORMAT_IS_MIN_SIZE(dfmt) (!GET_DISPLAY_FORMAT_IS_FIXED_SIZE(dfmt))

ht_vfs *compare_keys_direntry_cvfs;

int compare_keys_direntry(ht_data *key_a, ht_data *key_b)
{
	ht_data_direntry *a=(ht_data_direntry*)key_a;
	ht_data_direntry *b=(ht_data_direntry*)key_b;
	int a_dir=(HT_S_ISDIR(a->stat.mode)!=0);
	int b_dir=(HT_S_ISDIR(b->stat.mode)!=0);
	if (a_dir==b_dir) {
		return compare_keys_direntry_cvfs->filename_compare(a->name, b->name);
	}
	return b_dir-a_dir;
}

void ht_vfs_sub::init(ht_list *_vfss, char *starturl)
{
	ht_sub::init(0);
	set_display_format("type,name:20+,|,desc,|,bsize,|,perm,|,rmtime");
	dir=new ht_clist();
	((ht_clist*)dir)->init(compare_keys_direntry);
	dirsort=0;
	vfss=_vfss;
	cvfs=0;
	cproto=0;
	cwd=0;
	churl(starturl);
}

void ht_vfs_sub::done()
{
	if (cwd) free(cwd);
	if (cproto) free(cproto);
	if (dirsort) free(dirsort);
	dir->destroy();
	delete dir;
	ht_sub::done();
}

bool ht_vfs_sub::chdir(char *dir)
{
	char dirname[256];	/* FIXME: possible buffer overflow ! */
	if (cvfs->canonicalize(dir, dirname)==0) {
		if (cwd) free(cwd);
		cwd=strdup(dirname);
		refreshdir();
		if (uformat_viewer) uformat_viewer->goto_offset(0);
		return 1;
	}
	return 0;
}

bool ht_vfs_sub::churl(char *url)
{
/* extract protocol identifier if available */
	char *p=url;
	while (isalnum(*p) && *p) p++;
	if (cproto) free(cproto);
	if (strncmp(p, "://", 3)==0) {
		cproto=(char*)malloc(p-url+1);
		memmove(cproto, url, p-url);
		cproto[p-url]=0;
		url+=p-url+3;
	} else {
		cproto=(char*)malloc(1);
		*cproto=0;
	}
/* search a fitting vfs */
	cvfs=0;
	ht_vfs *vfs=(ht_vfs*)vfss->enum_first();
	while (vfs) {
		if (vfs->set_protocol(cproto)) {
			cvfs=vfs;
			break;
		}
		vfs=(ht_vfs*)vfss->enum_next();
	}
	if (!cvfs) return 0;
/* change current dir */
	return chdir(url);
}

bool ht_vfs_sub::convert_ofs_to_id(FILEOFS offset, ID *id1, ID *id2)
{
	if (offset<dir->count()) {
		*id1=offset;
		*id2=0;
		return 1;
	}
	return 0;
}

void ht_vfs_sub::first_line_id(ID *id1, ID *id2)
{
	*id1=0;
	*id2=0;
}

bool ht_vfs_sub::getline(char *line, ID id1, ID id2)
{
	ht_data_direntry *e=(ht_data_direntry*)dir->get(sortidx(id1));
	if (e) {
		for (int i=0; i<display_format_length; i++) {
			int timei=0;
			bool alignright=1;
			bool link_it=0;
			char *start=line;
			switch (GET_DISPLAY_FORMAT_TYPE(display_format[i])) {
				case VFSV_FORMAT_NAME:
					strcpy(line, e->name);
					line+=strlen(e->name);
					alignright=0;
					link_it=1;
					break;
				case VFSV_FORMAT_SIZE:
					if (e->stat.caps & pstat_size) {
						line+=sprintf(line, "%u", e->stat.size);
					}
					break;
				case VFSV_FORMAT_BSIZE:
					if (e->stat.caps & pstat_size) {
						if (HT_S_ISDIR(e->stat.mode)) {
							if (strcmp(e->name, "..")==0) {
								strcpy(line, "<UP-DIR>");
								line+=8;
							} else {
								strcpy(line, "<SUB-DIR>");
								line+=9;
							}
						} else {
							line+=sprintf(line, "%u", e->stat.size);
						}
					}
					break;
				case VFSV_FORMAT_TYPE:
					if (e->stat.caps & pstat_mode_type) {
						if (HT_S_ISDIR(e->stat.mode)) {
							*(line++)='/';
						} else if (HT_S_ISBLK(e->stat.mode)) {
							*(line++)='+';
						} else if (HT_S_ISCHR(e->stat.mode)) {
							*(line++)='#';
						} else if (HT_S_ISFIFO(e->stat.mode)) {
							*(line++)='|';
						} else if (HT_S_ISLNK(e->stat.mode)) {
							*(line++)='@';
						} else if (HT_S_ISSOCK(e->stat.mode)) {
							*(line++)='=';
						} else if (e->stat.mode & HT_S_IXUSR) {
							*(line++)='*';
						} else {
							*(line++)=' ';
						}
					}
					break;
				case VFSV_FORMAT_MTIME:timei++;
				case VFSV_FORMAT_ATIME:timei++;
				case VFSV_FORMAT_CTIME:timei++;
				case VFSV_FORMAT_RMTIME:timei++;
				case VFSV_FORMAT_RATIME:timei++;
				case VFSV_FORMAT_RCTIME: {
					time_t q;
					bool avail=0;
								bool reltime=0;
								
					switch (timei) {
						case 2:	reltime=1;
						case 5:
							if (e->stat.caps & pstat_mtime) {
								q=e->stat.mtime;
								avail=1;
							}
							break;
						case 1:	reltime=1;
						case 4:
							if (e->stat.caps & pstat_atime) {
								q=e->stat.atime;
								avail=1;
							}
							break;
						case 0:	reltime=1;
						case 3:
							if (e->stat.caps & pstat_ctime) {
								q=e->stat.ctime;
								avail=1;
							}
							break;
					}
					if (avail) {
						tm t=*gmtime(&q);

						time_t ct;
						time(&ct);
						tm c=*gmtime(&ct);
									   
						char *months[12]={"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
						if ((reltime) && ((UINT)ct-q<=60*60*24*28)) {
							if ((UINT)ct-q<=60*60*24) {
								line+=sprintf(line, "now   ");
							} else if ((UINT)ct-q<=60*60*24*2) {
								line+=sprintf(line, "yestrd");
							} else {
								line+=sprintf(line, "%s-%uday", ((UINT)(ct-q)/(60*60*24)>10) ? "" : " ", (UINT)(ct-q)/(60*60*24));
							}
						} else {
							line+=sprintf(line, "%s %02d", months[t.tm_mon], t.tm_mday);
						}
						if (t.tm_year==c.tm_year) {
							if ((reltime) && ((UINT)ct-q<=60*60*24)) {
								if ((UINT)ct-q<=60*60) {
									line+=sprintf(line, "-%umin", (UINT)(ct-q)/60);
								} else {
									line+=sprintf(line, "-%um:%u", (UINT)(ct-q)/60/60, (UINT)(ct-q)/60%60);
								}
							} else {
								line+=sprintf(line, " %02d:%02d", t.tm_hour, t.tm_min);
							}
						} else {
							line+=sprintf(line, "  %04d", t.tm_year+1900);
						}
					}
					break;
				}
				case VFSV_FORMAT_PERM:
					if (e->stat.caps & pstat_mode_type) {
						*(line++)=HT_S_ISDIR(e->stat.mode) ? 'd' : '-';
					}
					
					if (e->stat.caps & pstat_mode_oth) {
						*(line++)=(e->stat.mode & HT_S_IROTH) ? 'r' : '-';
						*(line++)=(e->stat.mode & HT_S_IWOTH) ? 'w' : '-';
						*(line++)=(e->stat.mode & HT_S_IXOTH) ? 'x' : '-';
					}
					
					if (e->stat.caps & pstat_mode_grp) {
						*(line++)=(e->stat.mode & HT_S_IRGRP) ? 'r' : '-';
						*(line++)=(e->stat.mode & HT_S_IWGRP) ? 'w' : '-';
						*(line++)=(e->stat.mode & HT_S_IXGRP) ? 'x' : '-';
					}
					
					if (e->stat.caps & pstat_mode_usr) {
						*(line++)=(e->stat.mode & HT_S_IRUSR) ? 'r' : '-';
						*(line++)=(e->stat.mode & HT_S_IWUSR) ? 'w' : '-';
						*(line++)=(e->stat.mode & HT_S_IXUSR) ? 'x' : '-';
					}
					break;
				case VFSV_FORMAT_MODE:
					if (e->stat.caps & pstat_mode_all) {
						line+=sprintf(line, "%o", e->stat.mode & ((1<<9)-1));
					}
					break;
				case VFSV_FORMAT_NLINK:
					break;
				case VFSV_FORMAT_NGID:
					if (e->stat.caps & pstat_gid) {
						line+=sprintf(line, "%u", e->stat.gid);
					}
					break;
				case VFSV_FORMAT_NUID:
					if (e->stat.caps & pstat_uid) {
						line+=sprintf(line, "%u", e->stat.uid);
					}
					break;
				case VFSV_FORMAT_OWNER:
					break;
				case VFSV_FORMAT_GROUP:
					break;
				case VFSV_FORMAT_INODE:
					if (e->stat.caps & pstat_inode) {
						line+=sprintf(line, "%u", e->stat.inode);
					}
					break;
				case VFSV_FORMAT_SPACE:
					*(line++)=' ';
					break;
				case VFSV_FORMAT_MARK:
					break;
				case VFSV_FORMAT_SEPARATOR:
					*(line++)='|';
					break;
				case VFSV_FORMAT_DESC:
					if (e->stat.caps & pstat_desc) {
						tag_strcpy(line, e->stat.desc);
						line+=tag_strlen(line);
						alignright=0;
					}
					break;
			}

			*line=0;
			int vvsize=tag_strvlen(start);
			int msize=GET_DISPLAY_FORMAT_SIZE(display_format[i]);
			if (GET_DISPLAY_FORMAT_IS_FIXED_SIZE(display_format[i])) {
				int m=line-start;
				if (m>msize) m=msize;
				if (alignright) {
					memmove(start+msize-m, line-m, m);
					memset(start, ' ', msize-m);
				} else {
					if (msize>line-start) memset(line, ' ', msize-(line-start));
				}
			} else {
				int csize=line-start;
				if (csize>msize) msize=csize;
				int vsize=max_prop_width[GET_DISPLAY_FORMAT_TYPE(display_format[i])]+(line-start)-vvsize;
				if (vsize>msize) msize=vsize;
				if (alignright) {
					memmove(start+msize-(line-start), start, line-start);
					if (msize>line-start) memset(start, ' ', msize-(line-start));
				} else {
					if (msize>line-start) memset(line, ' ', msize-(line-start));
				}
			}
			line=start+msize;
			if (link_it) {
				char p[512], *q;
				q=tag_make_ref_len(p, id1, 0, start, msize);
				memmove(start, p, q-p);
				line=start+(q-p);
			}
		}
		*line=0;
		return 1;
	}
	return 0;
}

void ht_vfs_sub::last_line_id(ID *id1, ID *id2)
{
	*id1=dir->count();
	*id2=0;
}

int ht_vfs_sub::next_line_id(ID *id1, ID *id2, int n)
{
	int r=n;
	int c=dir->count();
	ID i1=*id1;
	i1+=n;
	if ((int)i1>c-1) {
		r-=i1-c+1;
		i1=c-1;
	}
	if (r) *id1=i1;
	return r;
}

int ht_vfs_sub::prev_line_id(ID *id1, ID *id2, int n)
{
	int r;
	ID i1=*id1;
	if (i1<(dword)n) {
		r=i1;
		i1=0;
	} else {
		r=n;
		i1-=n;
	}
	if (r) *id1=i1;
	return r;
}

bool ht_vfs_sub::ref(ID id1, ID id2)
{
	ht_data_direntry *e=(ht_data_direntry*)dir->get(id1);
	if (e) {
		char d[256], f[256];	/* FIXME: possible buffer overflow ! */
		int dnl=strlen(cwd);
		strcpy(d, cwd);
		if ((d[dnl-1]!='\\') && (d[dnl-1]!='/')) {
			d[dnl]='/';
			d[dnl+1]=0;
		}
		strcat(d, e->name);
		if (HT_S_ISDIR(e->stat.mode)) {
			bool is_pp=(strcmp(e->name, "..")==0);
			f[0]=0;
			if (is_pp) {
				char *last_slash1 = strrchr(cwd, '/');
				char *last_slash2 = strrchr(cwd, '\\');
				char *last_slash = (last_slash1>last_slash2) ? last_slash1 : last_slash2;
				if (last_slash) {
					last_slash++;
					strcpy(f, last_slash);
				}
			}
			chdir(d);
			if (is_pp) {
				int i=0;
				ht_data_direntry *e=(ht_data_direntry*)dir->enum_first();
				while (e) {
					if (cvfs->filename_compare(e->name, f)==0) {
						if (uformat_viewer) uformat_viewer->goto_offset(i);
						break;
					}
					e=(ht_data_direntry*)dir->enum_next();
					i++;
				}
			}
		} else {
			cvfs->open(d, 1);
		}
	}
	return 0;
}

void ht_vfs_sub::refreshdir()
{
	dir->empty();
/* stat all entries & calculate max visual width of fields */
	for (int i=0; i<VFSV_FORMAT_PROPERTIES; i++) {
		max_prop_width[i]=0;
	}
/* read dir entries */
	pfind_t f;
	if (cvfs->findfirst(cwd, &f)) {
		do {
			if (strcmp(f.name, ".")!=0) {
				ht_data_direntry *e=new ht_data_direntry();
				e->name=strdup(f.name);
				e->stat=f.stat;
				dir->insert(e);
				update_max_widths(e);
			}
		} while (cvfs->findnext(&f));
		cvfs->findclose(&f);
	}
/* sort entries */
	compare_keys_direntry_cvfs=cvfs;
	dir->sort();
}

void ht_vfs_sub::set_display_format(char *fmt)
{
	int type;
	display_format_length=0;
	while ((fmt=translate_prop(fmt, &type))) {
		if (*fmt==':') {
			fmt++;
			unsigned int width=strtoul(fmt, &fmt, 10);
			if (*fmt=='+') {
				fmt++;
				display_format[display_format_length]=MAKE_DISPLAY_FORMAT_MIN_WIDTH(type, width);
			} else {
				display_format[display_format_length]=MAKE_DISPLAY_FORMAT_FIXED_WIDTH(type, width);
			}
		} else {
			display_format[display_format_length]=MAKE_DISPLAY_FORMAT(type);
		}

		while ((*fmt==',') || (*fmt==' ')) fmt++;
		if (++display_format_length==VFSV_FORMAT_MAX_LENGTH) break;
	}
}

int ht_vfs_sub::sortidx(int idx)
{
	if (dirsort) return dirsort[idx]; else return idx;
}

char *ht_vfs_sub::translate_prop(char *fmt, int *type)
{
	for (int i=0; i<VFSV_FORMAT_PROPERTIES; i++) {
		int l=strlen(format_property[i]);
		if (strncmp(fmt, format_property[i], l)==0) {
			*type=i;
			return fmt+l;
		}
	}
	return 0;
}

void ht_vfs_sub::unlink(ID id1, ID id2)
{
	char filename[260];
	ht_data_direntry *e=(ht_data_direntry*)dir->get(id1);
	strcpy(filename, cwd);
	if (filename[strlen(filename)-1]!='/') strcat(filename, "/");
	strcat(filename, e->name);
	if (HT_S_ISREG(e->stat.mode)) {
		if (confirmbox("Really delete file '%s' ?", filename)==button_ok) {
			cvfs->delete_file(filename);
			refreshdir();
		}
	} /*else if (HT_S_ISDIR(e->stat.mode)) {
		if (confirmbox("Really delete directory '%s' ?", filename)==button_ok) {
			cvfs->delete_file(filename);
			refreshdir();
		}
	} */else {
		infobox("unlink() not support this type of file...");
	}
}

void ht_vfs_sub::mkdir(char *name)
{
	if (name) {
		char dirname[260];
		strcpy(dirname, cwd);
		strcat(dirname, "/");
		strcat(dirname, name);
		cvfs->mkdir(dirname);
		refreshdir();
	}
}


int intstrlen(int a)
{
	int l=a ? 1 : 0;
	while (a/=10) l++;
	return l;
}

void ht_vfs_sub::update_max_widths(ht_data_direntry *e)
{
#define SATURATE(a, b) { int c=b; if (c>a) a=c; }
	SATURATE(max_prop_width[VFSV_FORMAT_NAME], tag_strlen(e->name));
	if (e->stat.caps & pstat_size) SATURATE(max_prop_width[VFSV_FORMAT_SIZE], intstrlen(e->stat.size));
	if (e->stat.caps & pstat_mode_type) max_prop_width[VFSV_FORMAT_TYPE]=1;
	if (e->stat.caps & pstat_size) max_prop_width[VFSV_FORMAT_BSIZE]=9;	// FIXME: involve strings somehow ("<UP-DIR>", "<SUB-DIR>")
	   
	if (e->stat.caps & pstat_mtime) max_prop_width[VFSV_FORMAT_MTIME]=12;
	if (e->stat.caps & pstat_atime) max_prop_width[VFSV_FORMAT_ATIME]=12;
	if (e->stat.caps & pstat_ctime) max_prop_width[VFSV_FORMAT_CTIME]=12;
	if (e->stat.caps & pstat_mtime) max_prop_width[VFSV_FORMAT_RMTIME]=12;
	if (e->stat.caps & pstat_atime) max_prop_width[VFSV_FORMAT_RATIME]=12;
	if (e->stat.caps & pstat_ctime) max_prop_width[VFSV_FORMAT_RCTIME]=12;
	
	max_prop_width[VFSV_FORMAT_PERM]=0;
	if (e->stat.caps & pstat_mode_type) max_prop_width[VFSV_FORMAT_PERM]++;
	if (e->stat.caps & pstat_mode_usr) max_prop_width[VFSV_FORMAT_PERM]+=3;
	if (e->stat.caps & pstat_mode_grp) max_prop_width[VFSV_FORMAT_PERM]+=3;
	if (e->stat.caps & pstat_mode_oth) max_prop_width[VFSV_FORMAT_PERM]+=3;
	
	if (e->stat.caps & pstat_mode_all) max_prop_width[VFSV_FORMAT_MODE]=3;
//	SATURATE(max_prop_width[VFSV_FORMAT_NLINK], intstrlen(e->stat.size));
	if (e->stat.caps & pstat_gid) SATURATE(max_prop_width[VFSV_FORMAT_NGID], intstrlen(e->stat.gid));
	if (e->stat.caps & pstat_uid) SATURATE(max_prop_width[VFSV_FORMAT_NUID], intstrlen(e->stat.uid));
// FIXME: owner...
// FIXME: group...
	if (e->stat.caps & pstat_inode) SATURATE(max_prop_width[VFSV_FORMAT_INODE], intstrlen(e->stat.inode));
// FIXME: space...
	max_prop_width[VFSV_FORMAT_MARK]=1;
	if (e->stat.caps & pstat_desc) SATURATE(max_prop_width[VFSV_FORMAT_DESC], tag_strvlen(e->stat.desc));
}

